from typing import Any
from typing import Dict
from typing import List

import dict_tools.data as data


async def profiles(
    hub,
    acct_file: str = None,
    acct_key: str = None,
    acct_data: Dict = None,
    **kwargs,
) -> Dict[str, List[Dict[str, Any]]]:
    """
    Read the acct information from the named subs and return the context
    """
    if acct_file and acct_data is None:
        hub.log.debug("Reading profiles from acct")
        acct_data = await hub.acct.init.profiles(acct_file, acct_key, **kwargs)

    if acct_data is None:
        acct_data = {}

    ret_data = data.NamespaceDict()

    for provider, profiles in acct_data.items():
        ret_data[provider] = []
        if not isinstance(profiles, list):
            profiles = [profiles]
        for profile in profiles:
            for name, info in profile.items():
                new_ctx = data.NamespaceDict({name: info})
                ret_data[provider].append(new_ctx)
                if provider in hub.acct:
                    # This profile needs to run through an acct plugin for processing
                    new_info = await hub.acct.init.process(
                        {provider}, {provider: new_ctx}
                    )
                    if new_info[provider]:
                        ret_data[provider].append(data.NamespaceDict({name: new_info}))

    return ret_data



import asyncio
import dataclasses
import queue
from typing import Any
from typing import Dict
from typing import List

import dict_tools.data as data


def __init__(hub):
    # Start off with a synchronous queue
    hub.evbus.SERIAL_PLUGIN = "json"
    hub.evbus.BUS = queue.Queue()


@dataclasses.dataclass
class Event:
    profile: str
    body: bytes


async def get(hub):
    bus = await hub.evbus.broker.init()
    return await bus.get()


async def init(hub):
    """
    Initialize the event bus queue in the current loop
    """
    # Once we are running in the context of asyncio, replace the synchronous queue with an async queue
    if isinstance(hub.evbus.BUS, queue.Queue):
        new_bus = asyncio.Queue()
        while not hub.evbus.BUS.empty():
            event = hub.evbus.BUS.get()
            new_bus.put_nowait(event)

        hub.evbus.BUS = new_bus
    return hub.evbus.BUS


async def put(hub, body: str, profile: str = "default"):
    """
    Put an event on the bus as a single data class
    :param hub:
    :param body: The message body that will be serialized and put on the bus
    :param profile: The ctx profile name that should be used for this message
    :param serial_plugin: The pop-serial plugin used to serialize data
    """
    serialized: bytes = hub.serial[hub.evbus.SERIAL_PLUGIN].dump(body)
    bus: asyncio.Queue = await hub.evbus.broker.init()
    await bus.put(Event(body=serialized, profile=profile))


def put_nowait(hub, body: str, profile: str = "default"):
    """
    Put an event on the bus as a single data class
    """
    serialized: bytes = hub.serial[hub.evbus.SERIAL_PLUGIN].dump(body)
    hub.evbus.BUS.put_nowait(Event(body=serialized, profile=profile))


async def propagate(
    hub,
    contexts: Dict[str, List[Dict[str, Any]]],
    event: Event,
):
    """
    Publish an event to all the ingress queues
    """
    match_plugin = contexts.get("match_plugin", hub.match.PLUGIN)
    coros = []
    for plugin in hub.ingress:
        # get a list of allowed acct plugins for this publisher
        acct_plugins = getattr(plugin, "ACCT", [])
        if not acct_plugins:
            raise NotImplementedError(
                f"No acct providers specified for ingress.{plugin.name}, specify ACCT providers.  I.E.\n"
                f"def __init__(hub):\n"
                f"    hub.ingress.{plugin.name}.ACCT = ['acceptable_acct_providers']\n"
            )

        for acct_plugin in acct_plugins:
            provider = contexts.get(acct_plugin, {})
            for profile in provider:
                for name in profile:
                    # Treat the profile name as a pattern, and see if the event profile name matches
                    if hub.match[match_plugin].find(name=event.profile, pattern=name):
                        ctx = data.NamespaceDict(acct=profile[name])
                        coros.append(plugin.publish(ctx=ctx, body=event.body))

    for task in asyncio.as_completed(coros):
        # Don't stop the listener for a failure
        try:
            await task
        except Exception as e:
            # Log at the lowest level in case log messages are being sent to a queue
            hub.log.trace(f"Event failed to propagate: {e.__class__.__name__}: {e}")



CLI_CONFIG = {
    "serial_plugin": {},
}
CONFIG = {
    "serial_plugin": {
        "help": "The plugin used to serialize data. Defaults to json",
        "default": "json",
    },
}
DYNE = {
    "log": ["log"],
    "evbus": ["evbus"],
    "ingress": ["ingress"],
    "match": ["match"],
}



import logging.handlers


def wrapper(hub, **kwargs):
    class AsyncQueueWrapper:
        def __init__(self):
            self.profile = kwargs.get("profile", "logger")

        def put_nowait(self, record: logging.LogRecord):
            hub.evbus.broker.put_nowait(
                profile=self.profile,
                body=dict(
                    message=record.getMessage(),
                    module=record.module,
                    level=record.levelname.strip(),
                    timestamp=str(record.asctime).strip(),
                ),
            )

    return AsyncQueueWrapper()


def setup(hub, conf):
    """
    Log to the ingress queue
    """
    root = logging.getLogger()

    raw_level = conf["log_level"].strip().lower()
    if raw_level.isdigit():
        hub.log.INT_LEVEL = int(raw_level)
    else:
        hub.log.INT_LEVEL = hub.log.LEVEL.get(raw_level, root.level)

    root.setLevel(hub.log.INT_LEVEL)
    cf = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    ch = logging.StreamHandler()
    ch.setLevel(hub.log.INT_LEVEL)
    ch.setFormatter(cf)
    root.addHandler(ch)

    ff = logging.Formatter(fmt=conf["log_fmt_logfile"], datefmt=conf["log_datefmt"])
    _, kwargs = hub.render.cli.args(conf["log_handler_options"])

    qh = logging.handlers.QueueHandler(hub.log.queue.wrapper(**kwargs))
    qh.setLevel(hub.log.INT_LEVEL)
    qh.setFormatter(ff)
    root.addHandler(qh)



import asyncio
from typing import Any
from typing import Dict
from typing import List

STOP_ITERATION = object()


def __init__(hub):
    hub.evbus.STARTED = False
    for dyne in ("acct", "ingress", "output", "serial", "match"):
        hub.pop.sub.add(dyne_name=dyne)


async def start(hub, contexts: Dict[str, List[Dict[str, Any]]]):
    hub.log.debug("Starting event bus")
    hub.evbus.STARTED = True

    # Initialize the queue in the current loop
    bus: asyncio.Queue = await hub.evbus.broker.init()

    while True:
        event = await bus.get()
        if event is STOP_ITERATION:
            hub.log.info("Event bus was forced to stop")
            return
        await hub.evbus.broker.propagate(contexts=contexts, event=event)


async def stop(hub):
    """
    Stop the event bus
    """
    hub.log.debug("Stopping the event bus")
    hub.evbus.BUS.put_nowait(STOP_ITERATION)


async def join(hub):
    while not hub.evbus.STARTED:
        await hub.pop.loop.sleep(0)



import asyncio

"""
To use the internal queues, specify profile names for which queues should be created

.. code-block:: sls

    internal:
      default:
      my_profile:
        routing_key: my_routing_key1
      my_other_profile:
        routing_key: my_routing_key2
"""


def __init__(hub):
    hub.ingress.internal.QUEUE = {}
    hub.ingress.internal.ACCT = ["internal"]


async def get(hub, routing_key: str = None):
    if routing_key not in hub.ingress.internal.QUEUE:
        hub.ingress.internal.QUEUE[routing_key] = asyncio.Queue()
    return hub.ingress.internal.QUEUE[routing_key]


async def publish(hub, ctx, body: bytes):
    # initialize the queue with the current loop
    channel = await hub.ingress.internal.get(routing_key=ctx.acct.get("routing_key"))
    await channel.put(body)
